/* Heltec Wireless Stick Lite LoRa Pump Controller with Command ACK
  
   Function:
   LoRa Pump Controller - HeltecESP32V3PumpEcho but now using formal ACK-Packet Echo response
   
   Description:
   1. Uses packet structures defined in the PacketHandler library
   2. Test sketch for [Pump] ACK protocol development
   3. Add interrupt processing to simulate manual pump override
   4. Add REL bit processing
  
    6 May 2023 v0.0.4  Packet Echo with ACK Flag
    9 May 2023 v0.0.5  Add interrupt for controller override
   10 May 2023 v0.0.6  Add REL bit processing
   18 May 2023 v0.0.7  Add REL/ACK buffer
   22 May 2023 v0.0.8  TX_State code rearrangement
   30 May 2023 v0.0.9  Update readBatteryVoltage()
   01 Jun 2023 v0.0.10 General housekeeping to get the various implementations [more or less] consistent
   03 Jun 2023 v0.0.11 Send pump status with sensor updates
   10 Jun 2023 v0.0.12 Add watchdog timer
   06 Jul 2023 v0.0.13 Ammend checkAckBuffer logic
    9 Jul 2023 v0.0.14 Randomise reportingPeriod to avoid Node 'update syncing'
   
   9 Jul 2023
   Digital Concepts
   www.digitalconcepts.net.au
*/

#include <esp_int_wdt.h>        // ESP32 watchdog timers
#include <esp_task_wdt.h>

#include "Arduino.h"
#include "Wire.h"               // Need this to define I2C bus
#include "LoRaWan_APP.h"        // Heltec LoRa protocol

#include <Ticker.h>             // Asynchronous timer

#include <LangloLoRa.h>         // [Langlo] LoRa configuration parameters
#include <PacketHandler.h>      // [Langlo] LoRa Packet management
#include <LinkedList.h>         // Linked List support for REL/ACK buffer
#include <WSLV3Pins.h>          // Heltec Wireless Stick Lite V3 pin definitions
#include <nodeV3Y1AConfig.h>    // Specific local Node configuration parameters

#define wakePin           A4a_A1
#define pump_Ctrl         GPIO48
#define wdtTimeout 5

struct softwareRevision {
  const uint8_t major;			// Major feature release
  const uint8_t minor;			// Minor feature enhancement release
  const uint8_t minimus;		// 'Bug fix' release
};

softwareRevision sketchRevision = {0,0,13};
bool softRev = true;

Ticker sensorTimer, resendTimer, pumpWatchDogTimer;

// The reporting period is randomised to help avoid syncing of Node updates
// Even if near synchronous updates from different Nodes don't collide,
// the Gateway cannot process multiple packets in quick succession

const int lowerReportingLimit = 55; // 55 sec
const int upperReportingLimit = 65; // 65 sec
int reportingPeriod; // Will be randomised between lowerReportingLimit and upperReportingLimit

uint16_t messageCounter = 0;
const int wdtInterval = 60;  // Seconds
unsigned long wdtLimit = 180000;  // Milliseconds
unsigned long lastMessageTime = 0;
bool pumpTimerActive = false;
bool turnPumpOff = true;

const int resendInterval = 1;
const int eventCycleTime = resendInterval * 1000;
const int resendLimit = 3;
int resendCount = 0;
bool ackReceived = false;
bool resendTimerActive = false;

struct bufferElement {
  PacketHandler packet;
  unsigned long startTime;
  int resendCount;
};
bufferElement element;
LinkedList<bufferElement> *ackBuffer = new LinkedList<bufferElement>();
int bufferSize;
bool checkAckBuffer = false;

bool ackRequired = true;
bool sendAck = false;
bool readSensors = false;

const int pumpOverridePin = USR_Button;  // GPIO0
bool togglePumpState = false;
uint8_t pumpId = 1;
PH_powerStatus pumpPowerState = PH_POWER_OFF;
PH_packetType ackPacketType, messageType;
PacketHandler inPacket, outPacket;

static RadioEvents_t RadioEvents;
bool lora_idle = true;
void OnTxDone( void );
void OnTxTimeout( void );
void OnRxDone( void );

typedef enum {
  RX_State,
  TX_State
} States_t;

States_t state;
int16_t rssi,rxSize;

// Now the other functions we need to set up our environment

void hard_restart() {
  esp_task_wdt_init(1,true);
  esp_task_wdt_add(NULL);
  while(true);
}

void setup() {
  Mcu.begin();
  reportingPeriod = random(lowerReportingLimit,upperReportingLimit);
  Serial.begin(115200);
  delay(500);
  Serial.print("[setup] ");
  Serial.print(myName);
  Serial.println(" LoRa V3 Pump Controller Test");
  if ( softRev ) {
    printSoftRev();
    softRev = false;
  }
  
  Serial.println("[setup] Commence Set-up...");
  
  pinMode(Builtin_LED,OUTPUT);
  pinMode(Vext_Ctrl,OUTPUT);
  digitalWrite(Vext_Ctrl,LOW);
  pinMode(wakePin,OUTPUT);                  // Activate battery monitor
  pinMode(pump_Ctrl,OUTPUT);

  pinMode(ADC_Ctrl,OUTPUT);
  pinMode(VBAT_Read,INPUT);
  adcAttachPin(VBAT_Read);
  analogReadResolution(12);
//  initialReadBatteryVoltage();

  //  Initialise LoRa

  Serial.println("[setup] Initialise LoRa...");
    
  // LoRa parameters defined in LangloLoRa.h

  RadioEvents.TxDone = OnTxDone;
  RadioEvents.TxTimeout = OnTxTimeout;
  RadioEvents.RxDone = OnRxDone;
  Radio.Init( &RadioEvents );
  
  Radio.SetChannel( RF_FREQUENCY );
  Radio.SetTxConfig( MODEM_LORA, TX_OUTPUT_POWER, 0, LORA_BANDWIDTH,
                                  LORA_SPREADING_FACTOR, LORA_CODINGRATE,
                                  LORA_PREAMBLE_LENGTH, LORA_FIX_LENGTH_PAYLOAD_ON,
                                  true, 0, 0, LORA_IQ_INVERSION_ON, 3000 );
  Radio.SetRxConfig( MODEM_LORA, LORA_BANDWIDTH, LORA_SPREADING_FACTOR,
                                  LORA_CODINGRATE, 0, LORA_PREAMBLE_LENGTH,
                                  LORA_SYMBOL_TIMEOUT, LORA_FIX_LENGTH_PAYLOAD_ON,
                                  0, true, 0, 0, LORA_IQ_INVERSION_ON, true );
  Radio.Sleep( );

  Serial.println("[setup] LoRa initialisation complete");
  Serial.println();

  // Interrupt to toggle pump state

  pinMode(pumpOverridePin,INPUT_PULLUP);
  attachInterrupt( pumpOverridePin, pumpStateControl, FALLING);

  // initialise the Packet object instances
  
  inPacket.begin();
  messageCounter++;
  outPacket.begin(gatewayMAC,myMAC,messageCounter);

   // Let the broker know there's been a reset
  
  outPacket.setPacketType(RESET);
  outPacket.setResetCode(0);
  outPacket.serialOut();
  digitalWrite(Builtin_LED,HIGH);
  sendMessage(outPacket);
  digitalWrite(Builtin_LED,LOW);
  // We need to wait a tick here to let the packet go out before doing anything more with the radio or the packet gets lost
  delay(100);

  // Initialise pump status

  messageCounter++;
  element.packet.begin(gatewayMAC,myMAC,messageCounter,PUMP);
  element.packet.setRelFlag();
  element.packet.setPumpId(pumpId);
  element.packet.setPumpState(PH_POWER_OFF);
  digitalWrite(pump_Ctrl,LOW);
  element.packet.serialOut();
  digitalWrite(Builtin_LED,HIGH);
  sendMessage(element.packet);
  digitalWrite(Builtin_LED,LOW);
  delay(100);
  // Add the packet to the buffer of packets awaiting ACK
  element.resendCount = 1;
  element.startTime = millis();
  ackBuffer->add(element);
  state = RX_State;

// Start the sensor timer

  sensorTimer.attach( reportingPeriod, setSensorFlag );

//  esp_task_wdt_init(wdtTimeout,true);
//  esp_task_wdt_add(NULL);

  Serial.println("[setup] Set-up Complete");
  Serial.println();
}
 
void loop() {
/*  
// Pat the dog...  
  if ( millis() - last >= 4000 ) {
    esp_task_wdt_reset();
    last = millis(); 
  }
 */ 
//  Now just loop around checking to see if there's anything to process
	switch ( state ) {
		case TX_State: {
//      Serial.println("[loop TX] Into TX mode...");

      // We have four possible tasks to perform
      // 1. Check to see if we need to acknowledge receipt of a packet. We need to do this first because the sender
      //    is waiting for our response;
      // 2. If we have sensor data to send, we need to send that;
      // 3. If a local pump state change has been triggered, we need to advertise the fact to the Gateway/MQTT broker.
      //    This is time-sensitive because we will be expecting an acknowledgement almost immediately. This conflicts with
      //    the similar requirements of ACK buffer processing (4.), so we can only perform one of these functions in any one cycle;
      // 4. If we have packets that require acknowledgement, and need to be resent, we send them (only one each cycle) last, because
      //    we need to transition immediately thereafter into RX_State to receive the acknowledgement.

      if ( sendAck ) {
        Serial.println("[loop TX sendAck] Send ACK...");
        inPacket.setDestinationMAC(inPacket.sourceMAC());
        inPacket.setSourceMAC(myMAC);
        inPacket.setAckFlag();
        inPacket.hexDump();
        inPacket.serialOut();
        digitalWrite(Builtin_LED,HIGH);
        sendMessage(inPacket);
        digitalWrite(Builtin_LED,LOW);
        sendAck = false;
        // We need to wait a tick here to let the packet go out before doing anything more with the radio
        delay(100);
      }

      if ( readSensors ) {
        uint16_t batteryVoltage = 0;
        outPacket.setSequenceNumber(messageCounter);
        batteryVoltage=readBatteryVoltage();
        Serial.print("[loop TX] Battery Voltage: ");
        Serial.print( batteryVoltage );
        Serial.println(" mV");
        outPacket.setPacketType(VOLTAGE);
        outPacket.setVoltage(batteryVoltage);
        outPacket.hexDump();
        outPacket.serialOut();
        digitalWrite(Builtin_LED,HIGH);
        sendMessage(outPacket);
        digitalWrite(Builtin_LED,LOW);
        // We need to wait a tick here to let the packet go out before continuing for two reasons:
        // 1. For some reason, if we're too quick to continue, the packet doesn't go out
        // 2. The Gateway can't handle two packets in quick succession
        delay(100);
        messageCounter++;                 // Increment message count
        /*
        outPacket.setPacketType(ATMOSPHERE);
        readBMESensor();
        outPacket.hexDump();
        outPacket.serialOut();
        digitalWrite(Builtin_LED,HIGH);
        sendMessage(outPacket);
        digitalWrite(Builtin_LED,LOW);
        // We need to wait a tick here to let the packet go out before doing anything more with the radio or the packet gets lost
        delay(100);
        messageCounter++;                 // Increment message count
        Serial.println("[loop TX] Reset the sensor timer");
        */
        // Report pump status
        outPacket.setSequenceNumber(messageCounter);
        outPacket.setPacketType(PUMP);
        outPacket.setPumpId(pumpId);
        outPacket.setPumpState(pumpPowerState);
        outPacket.hexDump();
        outPacket.serialOut();
        digitalWrite(Builtin_LED,HIGH);
        sendMessage(outPacket);
        digitalWrite(Builtin_LED,LOW);
        // We need to wait a tick here to let the packet go out before doing anything more with the radio or the packet gets lost
        delay(100);
        messageCounter++;                 // Increment message count
        readSensors = false;
        sensorTimer.once(reportingPeriod, setSensorFlag);     // Come around again when the timer fires
      }
      
      if ( togglePumpState ) {
        bufferElement element;
        element.packet.begin(gatewayMAC,myMAC,messageCounter,PUMP);
        element.packet.setRelFlag();
        element.packet.setPumpId(pumpId);
        if (( pumpPowerState == PH_POWER_ON ) || turnPumpOff ) {
          pumpPowerState = PH_POWER_OFF;
          digitalWrite(pump_Ctrl,LOW);
          if ( pumpTimerActive ) {
            pumpWatchDogTimer.detach();
            pumpTimerActive = false;
          }
        } else {
          pumpPowerState = PH_POWER_ON;
          digitalWrite(pump_Ctrl,HIGH);
        }
        element.packet.setPumpState(pumpPowerState);
        element.packet.serialOut();
        digitalWrite(Builtin_LED,HIGH);
        sendMessage(element.packet);
        digitalWrite(Builtin_LED,LOW);
        delay(100);
        messageCounter++;                 // Increment message count
        if ( !resendTimerActive ) {
          Serial.println("[loop TX togglePumpState] Start resendTimer...");
          resendTimer.once( resendInterval, resendFunction );
          resendTimerActive = true;
        }
        // Add the packet to the buffer of packets awaiting ACK
        element.resendCount = 1;
        element.startTime = millis();
        ackBuffer->add(element);
        togglePumpState = false;
        Serial.println("[loop TX] Increment Message Counter");
      } else if ( checkAckBuffer ) {
        Serial.println("[loop TX checkAckBuffer] Begin...");
        // If there is one, check the first element of the ACK buffer
        // We only do one resend per cycle so as not to block a response to the packet
        // that is sent
        bufferSize = ackBuffer->size();
        while ( bufferSize > 0 ) {
          Serial.print("[loop TX checkAckBuffer] ACK buffer non-empty [");
          Serial.print(bufferSize);
          Serial.println("]");

          unsigned long currentTime = millis();
          unsigned long timeInterval;

          Serial.println("[loop TX checkAckBuffer] Retrieve first element...");
          element = ackBuffer->get(0);
          timeInterval = currentTime - element.startTime;
          Serial.print("[loop TX checkAckBuffer] Current Time: ");
          Serial.println(currentTime);
          Serial.print("[loop TX checkAckBuffer] Start Time: ");
          Serial.println(element.startTime);
          Serial.print("[loop TX checkAckBuffer] Time Interval: ");
          Serial.println(timeInterval);
          Serial.print("[loop TX checkAckBuffer] eventCycleTime: ");
          Serial.println(eventCycleTime);
          if (timeInterval > eventCycleTime) {
            Serial.println("[loop TX checkAckBuffer] Resend packet... ");
            resendCount = ++element.resendCount;
            Serial.print("[loop TX checkAckBuffer] Resend count: ");
            Serial.println(resendCount);
            if ( resendCount > resendLimit ) {
              Serial.println("[loop TX checkAckBuffer] Resend Limit exceeded... ");
              Serial.println("[loop TX checkAckBuffer] Delete packet...");
              element = ackBuffer->remove(0);
              // May want to do something more at this point...

              bufferSize = ackBuffer->size();
              Serial.print("[loop TX checkAckBuffer] New Buffer Size: ");
              Serial.println(bufferSize);
            } else {
              // Move the element to the end of the queue
              Serial.print("[loop TX checkAckBuffer] Resend count: ");
              Serial.println(resendCount);
              Serial.println("[loop TX checkAckBuffer] Move element to end of queue...");
              element = ackBuffer->remove(0);
              element.startTime = currentTime;
              element.resendCount = resendCount;
              ackBuffer->add(element);
              Serial.println("[loop TX checkAckBuffer] Resending packet...");
              digitalWrite(Builtin_LED,HIGH);
              sendMessage(element.packet);
              digitalWrite(Builtin_LED,LOW);
              delay(100);
              Serial.print("[loop TX checkAckBuffer] Resend Count: ");
              Serial.println(element.resendCount);
              // We need to go see if there's a response to this transmission, so we'll
              // pretend the buffer's empty to get us out of the buffer processing loop
              bufferSize = 0;
            }
          } else {
            bufferSize--;
          }
          Serial.println();
        }
        // We will have exited the loop after resending just one packet, so we need to
        // check to see if there's still more in the buffer that we need to come back for       
        bufferSize = ackBuffer->size();
        if ( bufferSize > 0 ) {
          if ( !resendTimerActive ) {
            Serial.println("[loop TX bufferSize] Start resendTimer...");
            resendTimer.once( resendInterval,resendFunction );
            resendTimerActive = true;
          }
        } else {
          checkAckBuffer = false;
        }
        Serial.println("[loop TX checkAckBuffer] End");
      }

      Serial.println(); 
      state = RX_State;
      break;
    }
    case RX_State: {
//      Serial.println("[loop RX] Into RX mode");
      if ( lora_idle ) {
        lora_idle = false;
        Radio.Rx(0);
      }
      Radio.IrqProcess( );
      break;
    }
    default:
      state = RX_State;
      break;
  }
}

void OnTxDone( void ) {
//  Serial.println("[OnTxDone] TX done!");
  Radio.Sleep( );
  lora_idle = true;    
  state = RX_State;
}

void OnTxTimeout( void ) {
//  Serial.println("[OnTxTimeout] TX Timeout...");
  Radio.Sleep( );
  lora_idle = true;    
  state = RX_State;
}

void OnRxDone( uint8_t *payload, uint16_t size, int16_t rssi, int8_t snr ) {
  digitalWrite(Builtin_LED,HIGH);
  inPacket.setContent(payload,size);
  digitalWrite(Builtin_LED,LOW);
  Radio.Sleep( );
  lora_idle = true;    

  if (inPacket.destinationMAC() == myMAC) {
    uint8_t byteCount = inPacket.packetByteCount();

    Serial.println("[OnRxDone] Packet received");
/*
    Serial.println("[OnRxDone] Header content:");
    Serial.print("       DestinationMAC: ");
    Serial.printf("%02X", inPacket.destinationMAC());
    Serial.println();
    Serial.print("            SourceMAC: ");
    Serial.printf("%02X", inPacket.sourceMAC());
    Serial.println();
    Serial.print("           Sequence #: ");
    Serial.println(inPacket.sequenceNumber());
    Serial.print("             REL Flag: ");
    if ( inPacket.relFlag() ) {
      Serial.println("True");
    } else {
      Serial.println("False");
    }
    Serial.print("          Packet Type: ");
    Serial.println(inPacket.packetTypeDescription());
    Serial.print("             ACK Flag: ");
    if ( inPacket.ackFlag() ) {
      Serial.println("True");
    } else {
      Serial.println("False");
    }
    Serial.print("   Payload Byte Count: ");
    Serial.println(inPacket.payloadByteCount());
    Serial.println();
*/
      //  Verify the checksum before we go any further
    
    if (inPacket.verifyPayloadChecksum()) {

      Serial.println("[OnRxDone] Process the packet...");

      // Send relevant information to the Serial Monitor

      inPacket.hexDump();
      inPacket.serialOut();
      // print RSSI & SNR of packet
      Serial.print("[OnRxDone] RSSI ");
      Serial.print(String(rssi));
      Serial.print("   SNR ");
      Serial.println(String(snr));

      // and finally, respond when required...

      if ( inPacket.ackFlag() ) {
        if ( verifyAck(inPacket) ) {
          // Nothing more to do here at the moment
        }
      } else {
        switch (inPacket.packetType()) {
          case PUMP: {
            pumpPowerState = inPacket.pumpState();
            if ( pumpPowerState == PH_POWER_ON ) {
              digitalWrite(pump_Ctrl,HIGH);
              if ( !pumpTimerActive ) {
                pumpWatchDogTimer.attach(wdtInterval,pumpStateCheck);
                pumpTimerActive = true;
              }
              lastMessageTime = millis();
            } else {
              digitalWrite(pump_Ctrl,LOW);
              pumpWatchDogTimer.detach();
              pumpTimerActive = false;
            }
            break;
          }
          default: {
            Serial.println("[OnRxDone packetType Switch] Well, what am I supposed to do with this...?");
          }
        }
        if ( inPacket.relFlag() ) {
          sendAck = true;
          state = TX_State;
        }
      }
      Serial.println();
    } else {
      Serial.println("[OnRxDone] Invalid checksum, packet discarded");
      Serial.println();
      inPacket.erasePacketHeader();
    }
/*
    Serial.println("[receivePacket] Checksum validation disabled");
    Serial.println();
    return true;
*/
  } else {

    // No need to go any further
/*
    Serial.print("[OnRxDone] This appears to be for someone else--To: 0x");
    Serial.print(inPacket.destinationMAC(),HEX);
    Serial.print(" From: 0x");
    Serial.println(inPacket.sourceMAC(),HEX);
*/
    inPacket.erasePacketHeader();
  }
}

uint16_t readBatteryVoltage() {
// ADC resolution
  const int resolution = 12;
  const int adcMax = pow(2,resolution) - 1;
  const float adcMaxVoltage = 3.3;
// On-board voltage divider
  const int R1 = 390;
  const int R2 = 100;
// Calibration measurements
  const float measuredVoltage = 4.13;
  const float reportedVoltage = 3.91;
// Measured value multiplication factor
  const float factor = (adcMaxVoltage / adcMax) * ((R1 + R2)/(float)R2) * (measuredVoltage / reportedVoltage);
  digitalWrite(ADC_Ctrl,LOW);
  delay(100);
  int analogValue = analogRead(VBAT_Read);
  digitalWrite(ADC_Ctrl,HIGH);

  float floatVoltage = factor * analogValue;
  uint16_t voltage = (int)(floatVoltage * 1000.0);
/*
  Serial.print("[readBatteryVoltage]        ADC : ");
  Serial.println(analogValue);
  Serial.print("[readBatteryVoltage]      Float : ");
  Serial.println(floatVoltage,3);
  Serial.print("[readBatteryVoltage] milliVolts : ");
  Serial.println(voltage);
*/
  return voltage;
}

void readBMESensor() {

}

void resendFunction() {
  Serial.println("[resendFunction] Timer fired...");
  int bufferSize = ackBuffer->size();
  if ( bufferSize > 0 ) {
    Serial.println("[resendFunction] Check Buffer...");
    checkAckBuffer = true;
    state = TX_State;
  } else {
    Serial.println("[resendFunction] Buffer empty");
    checkAckBuffer = false;
  }
  resendTimerActive = false;
}

bool verifyAck(PacketHandler ackPacket) {
  // Search the buffer for a matching outstanding ACK
  int i = 0;
  int bufferSize = ackBuffer->size();
  bool found = false;
//  Serial.print("[verifyAck] Buffer size: ");
//  Serial.println(bufferSize);
  while ( i < bufferSize && !found ) {
//    Serial.print("[verifyAck] Checking buffer element: ");
//    Serial.println(i);
    element = ackBuffer->get(i);
    ackPacketType = ackPacket.packetType();
    if ( ackPacket.sourceMAC() == element.packet.destinationMAC() &&
        ackPacket.sequenceNumber() == element.packet.sequenceNumber() && 
        ackPacketType == element.packet.packetType() ) {
      Serial.println("[verifyAck] ACK verified");
      element = ackBuffer->remove(i);  // Crashes if we don't use the form of remove() that copies the deleted element (?)
      found = true;
    } else {
      Serial.println("[verifyAck] No match");
      Serial.print("[verifyAck] ackPacket.sourceMAC(): ");
      Serial.println(ackPacket.sourceMAC());
      Serial.print("[verifyAck] element.packet.destinationMAC(): ");
      Serial.println(element.packet.destinationMAC());
      Serial.print("[verifyAck] ackPacket.sequenceNumber(): ");
      Serial.println(ackPacket.sequenceNumber());
      Serial.print("[verifyAck] element.packet.sequenceNumber(): ");
      Serial.println(element.packet.sequenceNumber());
     }
     i++;
  }
  if ( !found ) {
    Serial.println("[verifyAck] ACK verification failed");
    ackPacket.serialOut();
  }
  return found;
}

void sendMessage(PacketHandler packet) {
//  Serial.println("[sendMessage] Finished Packet");
  Radio.Send(packet.byteStream(), packet.packetByteCount());
  Serial.println("[sendMessage] Packet Sent");
  Serial.println();
}

void setSensorFlag() {
  readSensors = true;
  state = TX_State;
}

void pumpStateControl() {
  togglePumpState = true;
  state = TX_State;
}

void pumpStateCheck() {
  unsigned long tickerTime = millis() - lastMessageTime;
  if ( tickerTime > wdtLimit ) {
    togglePumpState = true;
    turnPumpOff = true;
    state = TX_State;
  }
}

void printSoftRev() {
  String sketchRev = String(sketchRevision.major) + "." + String(sketchRevision.minor) + "." + String(sketchRevision.minimus);

  Serial.println("[printSoftRev] Print software revision data...");
  Serial.println("[printSoftRev] Sketch   " + sketchRev);
  Serial.println("[printSoftRev] Packet   " + inPacket.softwareRevision(PACKET_HANDLER));
  Serial.println("[printSoftRev] Node     " + inPacket.softwareRevision(NODE_HANDLER));
  Serial.println();
  softRev = false;
}